from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Optional

from . import public

@public
def total_size(
    obj: Hashable,
    handlers: Optional[Dict[Any, Callable[..., Any]]] = ...,
    verbose: bool = ...,
) -> int:
    ...



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

from . import public

TAIL_CALL: Tuple[()] = ...

@public
def async_tail_call(
    active: bool = ...,
) -> Annotated[Callable[..., Any], Annotated[Optional[Tuple[()]], '@tail_call()']]:
    ...



"""Utility functions for while loops"""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from contextlib import suppress
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Type

# pylint: disable=import-error
from . import public


@public
def while_raised(
    exc_types: Tuple[Type[Exception]],
    target: Callable[..., Any],
    /,
    *args: Any,
    implicit_break: bool = True,
) -> Optional[Any]:
    """Repeats a target function while suppressing exceptions provided.
    This is a convenience wrapper around :py:func:`contextlib.suppress`.
    :param exc_types: The exception types to suppress.
    :type  exc_types: Tuple[Type[Exception]]
    :param target: The function to repeat.
    :type  target: Callable
    :param implicit_break: Whether to implicitly break out of the loop.
    :type  implicit_break: bool, defaults to True
    :return: The return value of the target function.
    """
    while True:
        with suppress(*exc_types):
            if implicit_break:
                return target(*args)
            target(*args)
            break
    return None



"""blastpipe: A utility library for modern Python."""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import sys
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version
from typing import Annotated
from typing import Any

try:
    __version__ = version('blastpipe')
except PackageNotFoundError:  # pragma: no cover
    pass


def public(obj: Any) -> Annotated[Any, '__all__.__contains__(__name__)']:
    """Declares an object as public."""
    mod = sys.modules[obj.__module__]
    # pylint: disable=unnecessary-dunder-call
    if hasattr(mod, '__all__'):
        mod.__all__.append(obj.__name__)
    else:
        mod.__setattr__('__all__', [obj.__name__])
    return obj



"""Sequence manipulation."""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from typing import Set

# pylint: disable=import-error
from . import public
from .tailcall import async_tail_call


async def _chr_union(*args: int) -> Set[str]:
    """creates a set of all characters in the union of the given chars"""
    chars = set()
    if len(args) == 2 and all(map(isinstance, args, (int,) * len(args))):
        start, stop = args
        chars |= {chr(i) for i in range(start, stop)}
        return chars
    for i in args:
        if isinstance(i, int):
            chars |= {chr(i)}
        elif len(i) == 2 and all(map(isinstance, i, (int,) * len(i))):
            return i
        else:
            raise TypeError(f'Expected Sequence or int, got {type(i)}')
    return chars


@public
def chr_union(*args: int) -> Set[str]:
    """Wraps an asynchronous sequence of int ranges or ints into a set of str
    :return: a set of chr()
       # 1. range(start=args[0], stop=args[1])
          * OR
       # 2. chr(arg[n] if type(arg[n])==int) | chr_union(arg[:2])
    :raises: TypeError
        if not a sequence of start,stop pairs and ints, an int, or a start, stop pair
    """
    return asyncio.run(async_tail_call()(_chr_union)(*args))



# noqa: INP001
"""Tail call optimization tests"""
# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import typing

import pytest
from hypothesis import given
from hypothesis import strategies as st

# pylint: disable=import-error
from blastpipe.sequence import chr_union
from blastpipe.tailcall import async_tail_call


@given(active=st.booleans())
def test_fuzz_async_tail_call(active: bool) -> None:
    """This test code was written by the `hypothesis.extra.ghostwriter` module"""
    async_tail_call(active=active)


def test_chr_union_range() -> None:
    """Test chr_union with a range"""
    assert chr_union((32, 35)) == {' ', '!', '"'}
    assert chr_union(32) == {' '}


@st.composite
def bad_input(draw: typing.Any) -> typing.Tuple[st.SearchStrategy[typing.Any], ...]:
    """Test bad input for chr_union"""
    return (
        draw(st.lists(st.floats())),
        draw(st.lists(st.text())),
        draw(st.lists(st.functions())),
        draw(st.lists(st.booleans())),
        draw(st.floats()),
        draw(st.text()),
        draw(st.functions()),
        draw(st.booleans()),
    )


@given(args=bad_input())
def test_fuzz_bad_input_chr_union(args: typing.Tuple[typing.Any]) -> None:
    """Test bad input for chr_union"""
    with pytest.raises(TypeError):
        _ = chr_union(args)



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Generic
from typing import TypeVar

from . import public

@public
class BaseMixin(ABC):

    @classmethod
    @abstractmethod
    def extend_with(cls: Any, instance: Any) -> Any:
        ...

_T = TypeVar('_T', bound=BaseMixin)

@public
class Mixin(Generic[_T], BaseMixin):  # type: ignore[misc]
    ...

@public
def mixin(cls: Any, base: Any) -> Any:
    ...



"""Backports starting from Python 3.9"""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import string
import sys
from typing import TYPE_CHECKING
from typing import Any
from typing import List
from typing import Optional
from typing import Protocol
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    import re

# pylint: disable=import-error
from . import public
from .mixin import Mixin
from .mixin import mixin


class IsTemplatePre311(Protocol):
    """Protocol to check if a Template is pre-3.11"""

    delimiter: str
    idpattern: str
    braceidpattern: Optional[str]
    flags: int
    pattern: re.Pattern[str]
    template: str

    def safe_substitute(self: Any) -> Optional[str]:
        """Abstract method to be implemented by base"""

    def substitute(self: Any) -> Optional[str]:
        """Abstract method to be implemented by base"""


_T = TypeVar('_T', bound='string.Template')


class TemplateGetIdentifiersMixin(Mixin[Any]):
    """Example Template mixin with preferred way of typing and using mixins"""

    delimiter: str
    idpattern: str
    braceidpattern: Optional[str]
    flags: int
    pattern: re.Pattern[str]
    template: str

    def safe_substitute(self: object) -> Optional[str]:
        """Abstract method to be implemented by base"""

    def substitute(self: object) -> Optional[str]:
        """Abstract method to be implemented by base"""

    def get_identifiers(self: IsTemplatePre311) -> List[str]:  # pragma: defer to string
        """Mixin method get_identifiers for older Template pre-3.11"""
        ids = []
        for i in self.pattern.finditer(self.template):
            named = i.group('named') or i.group('braced')
            if named is not None and named not in ids:
                ids.append(named)
            elif named is None and i.group('invalid') is None and i.group('escaped') is None:
                raise ValueError('Unrecognized named group in pattern', self.pattern)
        return ids

    @classmethod
    def extend_with(cls: type[IsTemplatePre311], instance: _T) -> _T:
        """Extend the class with the mixin instance."""
        instance.__class__ = type(
            f'{instance.__class__.__name__}With{cls.__name__}',
            (instance.__class__, cls),
            {},
        )
        return instance


# pylint: disable=line-too-long
if sys.version_info < (3, 11):  # pragma: defer to python
    Template = public(
        mixin(TemplateGetIdentifiersMixin, string.Template),
    )  # pragma: defer to string
else:
    Template = public(string.Template)  # pragma: defer to string



"""Mixin ABC, Generic, and helper function."""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Generic
from typing import TypeVar

# pylint: disable=import-error
from . import public


# pylint: disable=too-few-public-methods
@public
class BaseMixin(ABC):
    """Abstract mixin class"""

    @classmethod
    @abstractmethod
    def extend_with(cls: Any, instance: Any) -> Any:
        """extend the instance with the mixin cls"""


_T = TypeVar('_T', bound='BaseMixin')


@public
class Mixin(Generic[_T], BaseMixin):
    """Generic mixin class"""


@public
def mixin(cls: Any, base: Any) -> Any:
    """Helper function to extend the class with the base"""

    def __wrapper(*args: Any, **kwargs: Any) -> Any:
        """decorator"""
        return cls.extend_with(base(*args, **kwargs))

    return __wrapper



"""Utility for tail call optimization."""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import functools
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Tuple

# pylint: disable=import-error
from . import public

TAIL_CALL: Tuple[()] = ()


@public
def async_tail_call(
    active: bool = True,
) -> Annotated[Callable[..., Any], Annotated[Optional[Tuple[()]], '@tail_call()']]:
    """Async tail_call decorator.
    :param active: Whether to activate async tail call optimization.
    """

    def __wrapper(func: Callable[..., Any]) -> Optional[Callable[..., Any]]:
        """decorator"""

        async def __trampoline(*args: Any, **_: Mapping[Any, Any]) -> Tuple[Any, ...]:
            """Tail call optimization."""
            while args.__class__ is tuple:
                args = await func(*args)
            return args

        if active:
            functools.update_wrapper(__trampoline, func)
            return __trampoline
        return None

    return __wrapper



"""Utilities for writing and reading length-prefix framed messages.
Using length-prefixed framing makes it easier for the reader to determine the
boundaries of each message before passing it to msgspec to be decoded.
"""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import TYPE_CHECKING  # pragma: no cover

# pylint: disable=import-error
from . import public  # pragma: no cover

if TYPE_CHECKING:  # pragma: no cover
    import asyncio  # pragma: defer to asyncio


# pylint: disable=used-before-assignment
@public  # pragma: defer to python
async def prefixed_send(stream: asyncio.StreamWriter, buffer: bytes) -> None:
    """Write a length-prefixed buffer to the stream"""
    prefix = len(buffer).to_bytes(4, 'big')
    stream.write(prefix)
    await stream.drain()
    stream.write(buffer)
    await stream.drain()


@public  # pragma: defer to python
async def prefixed_recv(stream: asyncio.StreamReader) -> bytes:
    """Read a length-prefixed buffer from the stream"""
    prefix = await stream.readexactly(4)
    # pylint: disable=invalid-name
    n = int.from_bytes(prefix, 'big')
    return await stream.readexactly(n)




import re
import string
from typing import Any
from typing import List
from typing import Optional
from typing import Protocol
from typing import TypeVar

from .mixin import Mixin

class IsTemplatePre311(Protocol):

    delimiter: str
    idpattern: str
    braceidpattern: Optional[str]
    flags: int
    pattern: re.Pattern[str]
    template: str
    def safe_substitute(self: Any) -> Optional[str]:
        ...
    def substitute(self: Any) -> Optional[str]:
        ...

_T = TypeVar('_T', bound=string.Template)

class TemplateGetIdentifiersMixin(Mixin[Any]):

    delimiter: str
    idpattern: str
    braceidpattern: Optional[str]
    flags: int
    pattern: re.Pattern[str]
    template: str
    def safe_substitute(self: object) -> Optional[str]:
        ...
    def substitute(self: object) -> Optional[str]:
        ...
    def get_identifiers(self: IsTemplatePre311) -> List[str]:
        ...
    @classmethod
    def extend_with(cls: type[IsTemplatePre311], instance: _T) -> _T:
        ...



"""Utilities for measuring memory usage."""

# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections import deque
from itertools import chain
from sys import getsizeof  # type: ignore
from sys import stderr
from typing import Any
from typing import Callable
from typing import Hashable
from typing import Iterable
from typing import Optional
from typing import Sized
from typing import Union

try:
    from reprlib import repr  # pylint: disable=redefined-builtin
except ImportError:  # pragma: defer to python
    pass

# pylint: disable=import-error
from . import public


@public
def total_size(  # noqa: C901
    obj: Hashable,
    handlers: Optional[dict[Any, Callable[..., Any]]] = None,
    verbose: bool = False,
) -> int:
    """Returns the approximate memory footprint an object and all of its contents.

    See `recipe <https://code.activestate.com/recipes/577504/>`__ for original.

    :param obj: object to measure size of
    :type  obj: object
    :param handlers: handlers to iterate over contents of containers, defaults to None
    :type  handlers: dict, optional
    :param verbose: whether to print verbose output, defaults to False
    :type  verbose: bool, optional
    :return: approximate memory footprint in bytes
    :rtype:  int

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}
    """
    if handlers is None:
        handlers = {}

    def dict_handler(
        _dict: dict[Any, Any],
    ) -> chain[Iterable[Any]]:  # pragma: defer to python
        """Default handler for dict objects"""
        return chain.from_iterable(_dict.items())

    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)
    seen = set()
    default_size = getsizeof(0)

    def sizeof(obj: Union[Sized, Hashable]) -> int:
        """return size of object in bytes"""
        if id(obj) in seen:  # pragma: defer to python
            return 0
        seen.add(id(obj))
        size_total = getsizeof(obj, default_size)

        if verbose:
            print(size_total, type(obj), repr(obj), file=stderr)  # type: ignore

        for typ, handler in all_handlers.items():
            if isinstance(obj, typ):
                size_total += sum(map(sizeof, handler(obj)))  # type: ignore
                break
        return size_total

    return sizeof(obj)



import asyncio

from . import public

@public
async def prefixed_send(stream: asyncio.StreamWriter, buffer: bytes) -> None:
    ...

@public
async def prefixed_recv(stream: asyncio.StreamReader) -> bytes:
    ...



# noqa: INP001
"""Mixin module tests."""
# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import typing

from hypothesis import given
from hypothesis import strategies as st

import blastpipe.loop
import blastpipe.malloc

# pylint: disable=import-error
import blastpipe.mixin


@st.composite
def sized_objects(
    draw: typing.Any,
) -> typing.Tuple[st.SearchStrategy[typing.Any], ...]:
    """sized object strategy"""
    return (
        draw(st.binary()),
        draw(st.booleans()),
        draw(st.complex_numbers()),
        draw(st.datetimes()),
        draw(st.floats()),
        draw(st.integers()),
        draw(st.text()),
        draw(st.timedeltas()),
        draw(st.timezone_keys()),
        draw(st.times()),
        draw(st.uuids()),
    )


# pylint: disable=invalid-name
@given(instance=st.from_type(object))
def test_fuzz_BaseMixin_extend_with(instance: typing.Any) -> None:
    """This test code was written by the `hypothesis.extra.ghostwriter` module"""
    blastpipe.mixin.BaseMixin.extend_with(instance=instance)


@given(cls=st.from_type(object), base=st.from_type(object))
def test_fuzz_mixin(cls: typing.Any, base: typing.Any) -> None:
    """This test code was written by the `hypothesis.extra.ghostwriter` module"""
    blastpipe.mixin.mixin(cls=cls, base=base)


@given(
    exc_types=st.tuples(st.just(Exception)),
    target=st.functions(),
    implicit_break=st.booleans(),
)
def test_fuzz_while_raised(
    exc_types: typing.Tuple[typing.Type[Exception]],
    target: typing.Callable[..., typing.Any],
    implicit_break: bool,
) -> None:
    """This test code was written by the `hypothesis.extra.ghostwriter` module"""
    blastpipe.loop.while_raised(exc_types, target, implicit_break=implicit_break)


@given(obj=sized_objects(), verbose=st.booleans())
def test_fuzz_total_size(obj: typing.Any, verbose: bool) -> None:
    """This test code was written by the `hypothesis.extra.ghostwriter` module"""
    blastpipe.malloc.total_size(obj=obj, handlers=None, verbose=verbose)



from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Type

from . import public

@public
def while_raised(
    exc_types: Tuple[Type[Exception]],
    target: Callable[..., Any],
    /,
    *args: Any,
    implicit_break: bool = ...,
) -> Optional[Any]:
    ...



# noqa: INP001
"""Tests for backported code."""
# Copyright 2023 Ross J. Duff MSc
# The copyright holder licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import string

from hypothesis import given
from hypothesis import strategies as st

# pylint: disable=import-error
import blastpipe.backports


# pylint: disable=invalid-name
@given(instance=st.builds(string.Template, st.text()))
def test_fuzz_TemplateGetIdentifiersMixin_extend_with(instance: string.Template) -> None:
    """Test extension of arbitrary :py:class:`string.Template` text.

    :param instance: Text Template
    :type instance: string.Template
    """
    blastpipe.backports.TemplateGetIdentifiersMixin.extend_with(instance=instance)
